Add-Type -AssemblyName System.Windows.Forms
Add-Type -AssemblyName System.Drawing

function Format-NumberWithMetricPrefix {
    param(
        [double]$Value,
        [int]$DecimalPlaces = 3 # How many decimal places to show after scaling
    )

    if ($Value -eq 0) { return "0" }

    $prefixes = @(
        @{ Limit = 1e12; Symbol = "T" } # Tera
        @{ Limit = 1e9;  Symbol = "G" } # Giga
        @{ Limit = 1e6;  Symbol = "M" } # Mega
        @{ Limit = 1e3;  Symbol = "k" } # kilo
        @{ Limit = 1e-2; Symbol = ""  } # Base range (>= 0.01 and < 1000)
        @{ Limit = 1e-3; Symbol = "m" } # milli
        @{ Limit = 1e-6; Symbol = [char]0x00B5 } # micro
        @{ Limit = 1e-9; Symbol = "n" } # nano
        @{ Limit = 1e-12; Symbol = "p" } # pico
        @{ Limit = 1e-15; Symbol = "f" } # femto
    )

    $absVal = [Math]::Abs($Value)
    $chosenPrefix = $null

    # Find prefix
    foreach ($prefixInfo in $prefixes) {
        # Use a tolerance for floating point comparison >= Limit
        if ($absVal -ge ($prefixInfo.Limit - 1e-18)) {
            $chosenPrefix = $prefixInfo
            break
        }
    }

    $scalingFactor = 1.0
    $prefixSymbol = ""

    if ($null -ne $chosenPrefix) {
        $prefixSymbol = $chosenPrefix.Symbol
        # Determine scaling factor based on symbol
        switch ($prefixSymbol) {
            "T" { $scalingFactor = 1e12 }
            "G" { $scalingFactor = 1e9 }
            "M" { $scalingFactor = 1e6 }
            "k" { $scalingFactor = 1e3 }
            ""  { $scalingFactor = 1.0 }
            "m" { $scalingFactor = 1e-3 }
            ([char]0x00B5) { $scalingFactor = 1e-6 } # Micro
            "n" { $scalingFactor = 1e-9 }
            "p" { $scalingFactor = 1e-12 }
            "f" { $scalingFactor = 1e-15 }
            default { $scalingFactor = 1.0 } # Should only happen for "" symbol
        }
        $scaledValue = $Value / $scalingFactor
        # Use "F" for fixed-point notation (e.g., F3 for 3 decimal places)
        $formatString = "{0:F$($DecimalPlaces)}"
        return "{0}{1}" -f ($formatString -f $scaledValue), $prefixSymbol

    } else {
        # Value is smaller than the smallest prefix limit (femto), use scientific notation
        # Adjust decimal places for scientific notation if needed
        $sciDecimalPlaces = $DecimalPlaces - 1
        if ($sciDecimalPlaces -lt 1) {$sciDecimalPlaces = 1}
        return "{0:E$($sciDecimalPlaces)}" -f $Value
    }
}
# --- End Helper Function ---

# --- GUI Elements ---
$form = New-Object System.Windows.Forms.Form
$form.Text = "FET Transfer Curve Analyzer by ripson"
# Initial size, might be adjusted later based on content
$form.ClientSize = New-Object System.Drawing.Size(580, 550) # Reduced initial height slightly
$form.StartPosition = [System.Windows.Forms.FormStartPosition]::CenterScreen
$form.AutoScaleMode = [System.Windows.Forms.AutoScaleMode]::Font
$form.Font = New-Object System.Drawing.Font("Segoe UI", 9) # Standard UI font

$dataLabel = New-Object System.Windows.Forms.Label
$dataLabel.Text = "Paste Transfer Curve Data (e.g., Vg Id):"
$dataLabel.Location = New-Object System.Drawing.Point(10, 10)
$dataLabel.AutoSize = $true

$dataTextBox = New-Object System.Windows.Forms.TextBox
$dataTextBox.Multiline = $true
$dataTextBox.ScrollBars = [System.Windows.Forms.ScrollBars]::Vertical
$dataTextBox.AcceptsReturn = $true
$dataTextBox.AcceptsTab = $true
$dataTextBox.WordWrap = $false
$dataTextBox.Font = New-Object System.Drawing.Font("Consolas", 9) # Monospace for data
$dataTextBox.Size = New-Object System.Drawing.Size(560, 180)
$dataTextBox.Location = New-Object System.Drawing.Point(10, 30)

$analyzeButton = New-Object System.Windows.Forms.Button
$analyzeButton.Text = "Analyze Curve"
$analyzeButton.Size = New-Object System.Drawing.Size(120, 30)
$buttonY = $dataTextBox.Bottom + 10 # Calculate Y coordinate first
$analyzeButton.Location = New-Object System.Drawing.Point(10, $buttonY) # Use the calculated value

# --- Output Fields ---
# Define positioning variables
$currentY = $analyzeButton.Bottom + 15 # Start 15px below the Analyze button's bottom edge
$labelX = 10
$valueX = 240
$valueWidth = 180
$verticalSpacing = 35 # Spacing between rows of Label/Textbox pairs

# Create Controls (On/Off Ratio Removed)
$linearityLabel = New-Object System.Windows.Forms.Label
$linearityLabel.Text = "Linearity (R" + [char]0x00B2 + ", full curve):"
$linearityLabel.AutoSize = $true
$linearityOutput = New-Object System.Windows.Forms.TextBox
$linearityOutput.ReadOnly = $true
$linearityOutput.Size = New-Object System.Drawing.Size($valueWidth, 20)

# OnOffRatioLabel and OnOffRatioOutput are REMOVED

$gmMaxLabel = New-Object System.Windows.Forms.Label
$gmMaxLabel.Text = "Max Transconductance (gm max) [S]:" 
$gmMaxLabel.AutoSize = $true
$gmMaxOutput = New-Object System.Windows.Forms.TextBox
$gmMaxOutput.ReadOnly = $true
$gmMaxOutput.Size = New-Object System.Drawing.Size($valueWidth, 20)

$gmAvgLabel = New-Object System.Windows.Forms.Label
$gmAvgLabel.Text = "Average Transconductance (gm avg) [S]:" 
$gmAvgLabel.AutoSize = $true
$gmAvgOutput = New-Object System.Windows.Forms.TextBox
$gmAvgOutput.ReadOnly = $true
$gmAvgOutput.Size = New-Object System.Drawing.Size($valueWidth, 20)

$vthLabel = New-Object System.Windows.Forms.Label
$vthLabel.Text = "Threshold Voltage (Vth, linear extr.) [V]:" 
$vthLabel.AutoSize = $true
$vthOutput = New-Object System.Windows.Forms.TextBox
$vthOutput.ReadOnly = $true
$vthOutput.Size = New-Object System.Drawing.Size($valueWidth, 20)

$gmLabel = New-Object System.Windows.Forms.Label
$gmLabel.Text = "Transconductance (gm = " + [char]0x0394 + "I / " + [char]0x0394 + "V) at Midpoint Voltage:"
$gmLabel.AutoSize = $true

$gmTextBox = New-Object System.Windows.Forms.TextBox
$gmTextBox.Multiline = $true
$gmTextBox.ReadOnly = $true
$gmTextBox.ScrollBars = [System.Windows.Forms.ScrollBars]::Vertical
$gmTextBox.Font = New-Object System.Drawing.Font("Consolas", 9) # Monospace font for alignment

# --- Position Output Controls Dynamically ---
# Row 1: Linearity
$labelY = $currentY
$textboxY = $currentY - 3
$linearityLabel.Location = New-Object System.Drawing.Point($labelX, $labelY)
$linearityOutput.Location = New-Object System.Drawing.Point($valueX, $textboxY)
$currentY += $verticalSpacing # Move down for the next row

# Row 2 (was 3): Max Transconductance
$labelY = $currentY
$textboxY = $currentY - 3
$gmMaxLabel.Location = New-Object System.Drawing.Point($labelX, $labelY)
$gmMaxOutput.Location = New-Object System.Drawing.Point($valueX, $textboxY)
$currentY += $verticalSpacing

# Row 3 (was 4): Average Transconductance
$labelY = $currentY
$textboxY = $currentY - 3
$gmAvgLabel.Location = New-Object System.Drawing.Point($labelX, $labelY)
$gmAvgOutput.Location = New-Object System.Drawing.Point($valueX, $textboxY)
$currentY += $verticalSpacing

# Row 4 (was 5): Threshold Voltage
$labelY = $currentY
$textboxY = $currentY - 3
$vthLabel.Location = New-Object System.Drawing.Point($labelX, $labelY)
$vthOutput.Location = New-Object System.Drawing.Point($valueX, $textboxY)
$currentY += $verticalSpacing

# Row 5 (was 6): Point-by-point Transconductance Label
$gmLabel.Location = New-Object System.Drawing.Point($labelX, $currentY)
$currentY += $gmLabel.Height + 5 # Add space below label based on its actual height

# Row 6 (was 7): Point-by-point Transconductance TextBox
$gmTextBox.Location = New-Object System.Drawing.Point($labelX, $currentY)
$gmTextBoxHeight = 150 # Desired height for this textbox
$gmTextBox.Size = New-Object System.Drawing.Size(560, $gmTextBoxHeight)

# --- Adjust Form Height ---
$formRequiredHeight = $gmTextBox.Bottom + 20 # Calculate required height + padding
$minHeight = $analyzeButton.Bottom + 220 # Reduced minimum height slightly
if ($formRequiredHeight -lt $minHeight) { $formRequiredHeight = $minHeight }
if ($formRequiredHeight -le 0) { $formRequiredHeight = 550 } # Safety check
$form.ClientSize = New-Object System.Drawing.Size($form.ClientSize.Width, $formRequiredHeight)

# Add controls to the form AFTER defining and positioning them
# *** REMOVED OnOffRatioLabel, OnOffRatioOutput from AddRange ***
$form.Controls.AddRange(@(
    $dataLabel, $dataTextBox, $analyzeButton,
    $linearityLabel, $linearityOutput,
    # $onOffRatioLabel, $onOffRatioOutput, # REMOVED
    $gmMaxLabel, $gmMaxOutput,
    $gmAvgLabel, $gmAvgOutput,
    $vthLabel, $vthOutput,
    $gmLabel, $gmTextBox
))

# --- Analysis Logic ---
$analyzeButton.Add_Click({
    # Clear previous results
    $linearityOutput.Text = ""
    $gmMaxOutput.Text = ""
    $gmAvgOutput.Text = ""
    $vthOutput.Text = ""
    $gmTextBox.Text = ""

    $rawData = $dataTextBox.Text.Trim()
    if ([string]::IsNullOrEmpty($rawData)) {
        [System.Windows.Forms.MessageBox]::Show("Please paste data into the text box.", "Input Error", [System.Windows.Forms.MessageBoxButtons]::OK, [System.Windows.Forms.MessageBoxIcon]::Warning)
        return
    }

    # --- Data Parsing (Unchanged) ---
    $lines = $rawData.Split([Environment]::NewLine, [System.StringSplitOptions]::RemoveEmptyEntries)
    $curveData = [System.Collections.Generic.List[PSCustomObject]]::new()
    $parseErrors = [System.Collections.Generic.List[string]]::new()
    foreach ($line in $lines) {
        $trimmedLine = $line.Trim()
        if ($trimmedLine -eq '' -or $trimmedLine.StartsWith('#') -or $trimmedLine.StartsWith('%') -or $trimmedLine -match '^[A-Za-z]') { continue }
        $parts = $null
        if ($trimmedLine.Contains("`t")) { $parts = $trimmedLine.Split("`t", [System.StringSplitOptions]::RemoveEmptyEntries) }
        elseif ($trimmedLine.Contains(" ")) { $parts = $trimmedLine.Split(" ", [System.StringSplitOptions]::RemoveEmptyEntries) }
        elseif ($trimmedLine.Contains(",")) { $parts = $trimmedLine.Split(",", [System.StringSplitOptions]::RemoveEmptyEntries) }
        else {
             if ($trimmedLine -match '^[+-]?\d+(\.\d+)?([Ee][+-]?\d+)?\s*[+-]?\d+(\.\d+)?([Ee][+-]?\d+)?$') {
                 $match = [regex]::Match($trimmedLine, '(?<=\d)(\s*[+-])')
                 if ($match.Success) {
                     $splitIndex = $match.Index
                     $parts = @($trimmedLine.Substring(0, $splitIndex).Trim(), $trimmedLine.Substring($splitIndex).Trim())
                 }
             }
             if ($null -eq $parts) { $parts = @($trimmedLine) }
        }
        if ($parts.Count -ge 2) {
            try {
                $voltage = [double]::Parse($parts[0].Trim(), [System.Globalization.CultureInfo]::InvariantCulture)
                $current = [double]::Parse($parts[1].Trim(), [System.Globalization.CultureInfo]::InvariantCulture)
                $curveData.Add([PSCustomObject]@{Voltage = $voltage; Current = $current})
            } catch { $parseErrors.Add("Could not parse line: '$line' - $($_.Exception.Message)") }
        } else { if ($trimmedLine){ $parseErrors.Add("Could not parse line (expected >= 2 columns): '$line'") } }
    }
    if ($parseErrors.Count -gt 0) {
        $maxErrorsToShow = 10
        $errorMsgItems = $parseErrors | Select-Object -First $maxErrorsToShow
        if ($parseErrors.Count -gt $maxErrorsToShow) { $errorMsgItems += "`n... ($($parseErrors.Count - $maxErrorsToShow) more errors)" }
        [System.Windows.Forms.MessageBox]::Show("Errors encountered during data parsing:`n`n$($errorMsgItems -join "`n")", "Parsing Error", [System.Windows.Forms.MessageBoxButtons]::OK, [System.Windows.Forms.MessageBoxIcon]::Warning)
        if ($curveData.Count -lt 2) { return }
    }
    if ($curveData.Count -lt 2) {
        $message = "Need at least 2 data points for analysis."
        # *** REMOVED onOffRatioOutput from controls list ***
        $controlsToUpdate = @($linearityOutput, $gmMaxOutput, $gmAvgOutput, $vthOutput, $gmTextBox)
        $controlsToUpdate.ForEach({ $_.Text = $message })
        [System.Windows.Forms.MessageBox]::Show($message, "Input Error", [System.Windows.Forms.MessageBoxButtons]::OK, [System.Windows.Forms.MessageBoxIcon]::Warning)
        return
    }
    # --- End Data Parsing ---

    # --- Perform Calculations ---

    # Calculate point-by-point Transconductance (gm)
    $gmResults = [System.Collections.Generic.List[PSCustomObject]]::new()
    $gmOutputText = [System.Collections.Generic.List[string]]::new()
    $gmOutputText.Add("Midpoint V [V]  Gm [S]            ") # Header
    $gmOutputText.Add("--------------  ------------------") # Underline

    for ($i = 1; $i -lt $curveData.Count; $i++) {
        $point1 = $curveData[$i-1]
        $point2 = $curveData[$i]
        $deltaV = $point2.Voltage - $point1.Voltage
        $deltaI = $point2.Current - $point1.Current

        if ($deltaV -ne 0) {
            $gmValue = $deltaI / $deltaV
            $midVoltage = ($point2.Voltage + $point1.Voltage) / 2
            $gmResults.Add([PSCustomObject]@{ MidVoltage = $midVoltage; Gm = $gmValue; Index = $i })
            $gmFormatted = Format-NumberWithMetricPrefix -Value $gmValue -DecimalPlaces 3
            $gmOutputText.Add(("{0,-15:N5} {1,-18}" -f $midVoltage, $gmFormatted).TrimEnd())
        } else {
             $midVoltage = ($point2.Voltage + $point1.Voltage) / 2
             $gmOutputText.Add(("{0,-15:N5} Vertical (Delta V=0)" -f $midVoltage).TrimEnd())
        }
    }
    $gmTextBox.Text = $gmOutputText -join "`r`n"

    if ($gmResults.Count -eq 0) {
         [System.Windows.Forms.MessageBox]::Show("Could not calculate any gm values (perhaps all Delta V were zero?). Analysis aborted.", "Calculation Error", [System.Windows.Forms.MessageBoxButtons]::OK, [System.Windows.Forms.MessageBoxIcon]::Warning)
         $controlsToUpdate = @($linearityOutput, $gmMaxOutput, $gmAvgOutput, $vthOutput)
         $controlsToUpdate.ForEach({ $_.Text = "Error" })
         return
    }

    # Calculate Max and Average gm
    $numericGmValues = $gmResults.Gm
    $gmStats = $numericGmValues | Measure-Object -Maximum -Average
    $gmMax = $gmStats.Maximum
    $gmAvg = $gmStats.Average
    $gmMaxOutput.Text = Format-NumberWithMetricPrefix -Value $gmMax -DecimalPlaces 3
    $gmAvgOutput.Text = Format-NumberWithMetricPrefix -Value $gmAvg -DecimalPlaces 3

    # Calculate Threshold Voltage (Vth)
    $vthValueString = "N/A"
    if ([Math]::Abs($gmMax) -gt 1e-18) {
        $gmMaxPointInfo = $null; foreach($item in $gmResults){if($item.Gm -eq $gmMax){$gmMaxPointInfo=$item; break}}
        if ($null -ne $gmMaxPointInfo) {
            $indexMaxGm = $gmMaxPointInfo.Index
            if ($indexMaxGm -ge 1 -and $indexMaxGm -lt $curveData.Count) {
                $point1_vth = $curveData[$indexMaxGm - 1]; $point2_vth = $curveData[$indexMaxGm]
                $interceptC = $point2_vth.Current - ($gmMax * $point2_vth.Voltage)
                try { $vthNumeric = -$interceptC / $gmMax; $vthValueString = "{0:N4}" -f $vthNumeric }
                catch { $vthValueString = "Calc Error" }
            } else { $vthValueString = "Index Error" }
        } else { $vthValueString = "Not Found" }
    } else { $vthValueString = "N/A (gmMax ~ 0)"}
    $vthOutput.Text = $vthValueString


    # Calculate linearity (R-squared)
    $n = $curveData.Count; $sumX = 0.0; $sumY = 0.0; $sumXY = 0.0; $sumX2 = 0.0; $sumY2 = 0.0
    foreach ($point in $curveData) { $x = $point.Voltage; $y = $point.Current; $sumX += $x; $sumY += $y; $sumXY += $x * $y; $sumX2 += $x * $x; $sumY2 += $y * $y }
    $denominatorTolerance = 1e-12; $denominatorX = ($n * $sumX2) - ($sumX * $sumX); $denominatorY = ($n * $sumY2) - ($sumY * $sumY)
    if ([Math]::Abs($denominatorX) -gt $denominatorTolerance -and [Math]::Abs($denominatorY) -gt $denominatorTolerance) {
        $numerator = ($n * $sumXY) - ($sumX * $sumY); $denomProduct = $denominatorX * $denominatorY
        if ($denomProduct -gt 0) { $rSquared = ($numerator * $numerator) / $denomProduct; if ($rSquared -lt 0) {$rSquared = 0}; if ($rSquared -gt 1) {$rSquared = 1}; $linearityOutput.Text = "{0:N4}" -f $rSquared }
        else { $linearityOutput.Text = "N/A (Data issue?)" }
    } else { $linearityOutput.Text = "N/A (flat data?)" }

}) # End of analyzeButton.Add_Click

# --- Show the Form ---
# $form.TopMost = $true # Optional: Keep window on top
$form.ShowDialog() | Out-Null

# Dispose of the form object when done
$form.Dispose()
